####
####  A growing set of functions to compute the 
####  intersections on boxes,
####  and take care of count objects.

#### Object GeneCount summarizes the read counts within
#### boxes of a gene.
#### this should be extended to any general object to account
#### for all other data types:
####  - Genes which are tail to tail.
####  - Paired-End reads,
####  - Allele Specific expression,
####  - etc...

####  and also extend to more complex data expression models (variance component
####  per bloc, or any type of summary statistic)


#################################################################
###
###   Count the reads intersecting with exons.
###

###A simple class creator
###

genecount.new <- function(name, counts, sj.counts = NULL){
  nex = length(counts)
  names(counts) = 1:nex
  tot.counts = sum(c(counts, sj.counts))
  if (length(sj.counts) >0){
    names(sj.counts) = 1:length(sj.counts)
  }
  res.count = list(
    name = name,
    counts.exons = counts,
    counts.junctions = sj.counts,
    counts.total = tot.counts
    )
  attr(res.count, 'class') <- "GeneCount"
  return(res.count)
}


####  From a gene object and intersections, get the counts
###   in each annotated exons,
####  For performance reason, the reads are supposed ordered
####  by default and in the same chromosome as the gene
####
####  The relative.pos parameter is meant to get a
####  percent of the sequence profile of hits within each exon.
####
CountReadsInExons <- function(gene, read.pos, read.length = 27, is.ordered = TRUE,
                              pos.col = "coord", count.col = "counts",
                              relative.pos = FALSE){
  #TODO: check the type of the gene object.
  if (is.null(gene$njunctions)){
    gene = CutFromConnexeJunction(gene, read.length = read.length)
  }
  counts.exons = rep(0, gene$nexons)
  names(counts.exons) = 1:gene$nexons
  counts.junctions = rep(0, gene$njunctions)
  if (length(counts.junctions)>0){
    names(counts.junctions) = 1:gene$njunctions
  }
  borders = range( c(gene$exons.pos$begin, gene$exons.pos$end - read.length + 1))
  pos2cons = which(read.pos[[pos.col]] >= borders[1] & read.pos[[pos.col]] <= borders[2])
  #
  if (length(pos2cons) == 0){
    gene.res = gene
    gene.res$counts.exons = counts.exons
    gene.res$counts.junctions = counts.junctions
    return(list(name = gene$name,
                counts.exons = counts.exons,
                counts.junctions = counts.junctions))
  }
  r.tmp = read.pos[pos2cons,]
  missed.pos = rep(TRUE, length(pos2cons))
  if (max(gene$exons.pos$length) >= read.length){
    list.counts.exons = CountInBoxes(r.tmp,
      gene$exons.pos$begin, gene$exons.pos$end,
      read.length = read.length, return.perc= relative.pos)
    ind.counts = match(names(list.counts.exons$count), names(counts.exons))
    counts.exons[ind.counts] = list.counts.exons$count
    missed.pos = list.counts.exons$missed
  }
  ##Now we test the junctions
  if (gene$njunctions > 0 & sum(missed.pos)>0 ){
    list.counts.junctions = CountInBoxes(r.tmp[missed.pos,],
      gene$junctions.coord$begin, gene$junctions.coord$end,
      read.length = read.length, overlap.boxes = TRUE, return.perc = relative.pos)
    counts.junctions = list.counts.junctions$count
  }
  #gene.res = gene
  #gene.res$counts.exons = counts.exons
  #gene.res$counts.junctions = counts.junctions
  pos.res = list(name = gene$name,
    counts.exons = counts.exons,
    counts.junctions = counts.junctions,
    counts.total = sum(c(counts.exons , counts.junctions)))
  if (relative.pos){
    pos.res$perc.exons = list.counts.exons$percents
  }
  attr(pos.res, 'class') <- "GeneCount"
  return(pos.res)
}

##
## Count the reads of length read.length starting in pos.tab 
## that are in boxes of coordinate begins-ends, supposed non overlapping
CountInBoxes <- function(pos.tab, begins, ends, pos.col = "coord",
                         count.col = "count",read.length = 27,
                         overlap.boxes = FALSE, return.perc = FALSE){
  c.ex = rep(0, length(begins))
  if (overlap.boxes){
    return(CountOverlapBoxes(pos.tab, begins, ends, pos.col = pos.col, count.col = count.col,
                             read.length = read.length))
  }
  true.ends = ends - read.length + 2
  bord = range(c(true.ends, begins))
  box.labels = 1:length(begins)
  beg.lim = which(max(true.ends) > begins)
  end.lim = which(min(begins) < true.ends)
  br.beg = c(begins[beg.lim], bord[2])
  br.end = c(bord[1], true.ends[end.lim])
  ##every read gets a factor for the box where it begins and the one
  ##where it ends
  box.gr.beg = cut(pos.tab[[pos.col]], br.beg,
    labels = box.labels[beg.lim], right = FALSE)
  box.gr.end = cut(pos.tab[[pos.col]], br.end,
    labels = box.labels[end.lim], right = FALSE)
  #all the reads position are tested to be a boxe
  in.boxes = as.character(box.gr.beg) == as.character(box.gr.end) & is.finite(box.gr.beg) & is.finite(box.gr.end)
  if (any(in.boxes)){
    c.ex = tapply(pos.tab[[count.col]][in.boxes], box.gr.beg[in.boxes], sum)
    c.ex[is.na(c.ex)] = 0
  }
  outside.box = !in.boxes
  result = list(count = c.ex, missed = outside.box)
  ##Add information about the percentages in each exons.
  ##A list with for each exon the percentage of hits over the distance.
  if (return.perc & any(in.boxes)){
    perc.by.ex = sapply(which(in.boxes), function(x){
      ex = as.numeric(as.character(box.gr.beg[x]))
      nr = pos.tab[[count.col]][x]
      ###A verifier
      c.beg = begins[ex]
      c.length = true.ends[ex] + 1 - c.beg
      perc.p = (pos.tab[[pos.col]][x]- c.beg)/(c.length)
      return(c(ex,perc.p,nr))
    })
    result$percents = t(perc.by.ex)
  }
  return(result)
}

####
####Count the positions falling in boxes with possible overlapping
CountOverlapBoxes <- function(pos.tab, begins, ends, pos.col = "coord",
                              count.col = "count", read.length = 27){
  true.ends = ends - read.length + 1
  posinbox <- function(pos){
    indexes = which(pos >= begins & pos <= true.ends)
    return(indexes)
  }
  l.p = sapply(pos.tab[[pos.col]], posinbox , simplify = FALSE)
  c.ex = rep(0, length(begins))
  names(c.ex) = 1:length(begins)
  outside.box = rep(FALSE, nrow(pos.tab))
  for (i in 1:nrow(pos.tab)){
    boxi = l.p[[i]]
    if (length(boxi)>0){
      c.ex[boxi] = c.ex[boxi] + pos.tab[[count.col]][i]
      outside.box[i] = TRUE
    }
  }
  result = list(counts = c.ex , missed = outside.box)
  return(result)
}


