###
###  Set of functions to handle a generic gene structure
###

###
###  TODO : check the last version for all functions interacting on the
###   
###


#########################################################
####
####            Creator functions
####
#########################################################

####
#### Gene Structure objects are defined with at least:
#### - name  (ID for the gene)
#### - names (IDs for the transcripts)
#### - chr (chromosome)
#### - direction +/-
#### - nforms : number of annotated isoforms
#### - nexons : total number of exons
#### - exons.pos : data frame of subexons coordinates
####      with at least columns : subexons, begin, end, length, dist.next
#### _ njunctions, which lists the number of connex junctions within annotation
#### (should have as well the read length information, needed at the construction
#### step of the connex junctions).

####
#### Simple constructor, the exons boundaries are supposed to allow
#### a unique cutting over all the transcripts.
####
gene.new <- function(info, indicator, exons.pos, read.length = 27){
  gene = list()
  nexons = info$nexons
  if (nexons != nrow(exons.pos)){
    warning("Problem with the number of exons lines wr to the gene info in", info$gene)
    nexons = nrow(exons.pos)
  }
  if (is.null(exons.pos$begin) | is.null(exons.pos$end)){
    stop("Error, exons positions not defined, Cannot find begin or end column")
  }
  
  gene$name = as.character(info$name)
  gene$chr = as.character(info$chr)
  gene$direction = as.character(info$direction)
  gene$nforms = info$nforms
  gene$names = as.character(indicator[,2])
  gene$nexons = nexons
  colkeep = which(!is.na(indicator[1,]))
  gene$I.big = indicator[,colkeep[-c(1,2)]]
  if (gene$nexons == 1){
    gene$I.big = data.frame(E1 = indicator[,colkeep[-c(1,2)]])
  }
  row.names(gene$I.big) = indicator[, 2]
  #order exons by genomic start positions
  ex.order = order(exons.pos$begin)
  #gene$exons.pos = exons.pos[ex.order,-1]
  gene$exons.pos = exons.pos[,-1]
  #add the ordering according to exon positions to indicator matrix and names
  #gene$I.big = gene$I.big[,ex.order]
  #names(gene$I.big) = names(gene$I.big)[ex.order]
  #gene$exons.pos$subexon = gene$exons.pos$subexon[ex.order]
  
  if (is.null(gene$exons.pos$length)){
    gene$exons.pos$length = gene$exons.pos$end - gene$exons.pos$begin + 1
  }
 
  if (gene$nexons == 1){
    gene$exons.pos$dist.next = NA
  }else{
    gene$exons.pos$dist.next = c(gene$exons.pos$begin[2:gene$nexons] -
      gene$exons.pos$end[1:(gene$nexons-1)], NA) ##We include the next one
    if (any(gene$exons.pos$dist.next < 0, na.rm = TRUE)){
      warning("Error in gene.new, some exons are overlapping")
    }
  }
  if (is.null(gene$exons.pos$subexon)){
    gene$exons.pos$subexon = paste("E", 1:nexons, sep = "")
  }
  gene = CutFromConnexeJunction(gene, read.length = read.length)
  attr(gene, 'class') <- "GeneStructure"
  return(gene)
}

###
### Utility function to generate the connex junction set
### from the exon annotation
###
CutFromConnexeJunction <- function(gene, read.length = 27){
  ind.con = gene$exons.pos$dist.next == 1
  if (!any(ind.con, na.rm = TRUE)){
    gene.res = gene
    gene.res$njunctions = 0
    return(gene.res)
  }
  e.t = paste("E", 1:gene$nexons, sep = "" )
  cols.n = na.omit(match(gene$exons.pos$subexon[ind.con], e.t))
  junctions.ind = sapply(cols.n, function(x)
    {return(as.numeric(apply(gene$I.big[,c(x,x+1)],1,sum)==2))})
  if (is.null(dim(junctions.ind))){
    junctions.ind = matrix(junctions.ind, ncol = length(junctions.ind))
  }
  diff = apply(junctions.ind,2, function(x) { any(x == 1)})
  junctions.ind = junctions.ind[,diff]
  cols.n = cols.n[diff]
  junctions = e.t[cols.n]
  ind.junc = gene$exons.pos$subexon %in% junctions
  ##### We just need to extend those one now.
  pos.2.extend = gene$exons.pos$end[ind.junc]
  junc.beg = pos.2.extend - read.length + 2
  junc.end = pos.2.extend + read.length
  ##### 
  junctions.pos = data.frame( e.beg = cols.n , e.end = cols.n+1,
    begin = junc.beg , end = junc.end)
  njunctions = nrow(junctions.pos)
  if (njunctions > 0){
    ###we need to verify that the borders of junctions are in exons
    for (i in 1:njunctions){
      ce.beg = junctions.pos$e.beg[i]
      ce.end = junctions.pos$e.end[i]
      ce.end = ce.end +
        max(which(ce.end + 0:(njunctions-i) == junctions.pos$e.end[i:njunctions])) - 1
      ce.beg = ce.beg -
          min(which((ce.beg - (0:(i-1))) == junctions.pos$e.beg[i:1])) + 1
      junctions.pos$begin[i] = max(gene$exons.pos$begin[ce.beg], junctions.pos$begin[i])
      junctions.pos$end[i] = min(gene$exons.pos$end[ce.end], junctions.pos$end[i])
    }
  }
  gene.res = gene
  junctions.pos$length = junctions.pos$end - junctions.pos$begin + 1
  gene.res$junctions.coord = junctions.pos
  gene.res$njunctions = njunctions
  gene.res$junctions.ind = junctions.ind
  attr(gene.res, 'class') <- "GeneStructure"
  return(gene.res)
}


#################################################################
###
###  Initialisation functions
###

###
### Generates a list with all the distant junctions used by every
###  transcript and their coordinates.
GenerateJunctionMatrix <- function(gene){
  ac =as.character
  I.mat = gene$I.big
  t.names = row.names(I.mat)
  junc.connex = c(na.omit(gene$exons.pos$dist.next) == 1, FALSE)
  list.juncs = list()
  for (i in 1:nrow(I.mat)){
    ###Problem here, we shift the detected event.
    cur.ex = which(I.mat[i,] == 1 & !junc.connex)
    ex.begs = cur.ex[1:(length(cur.ex)-1)]
    ex.ends = cur.ex[2:length(cur.ex)]
    list.juncs[[t.names[i]]]= data.frame(transcript = t.names[i],
                ID = paste("J", ex.begs, ex.ends, sep = "-"),
                e.beg = ex.begs, e.end = ex.ends,
                begin = gene$exons.pos$end[ex.begs],
                end = gene$exons.pos$begin[ex.ends])
  }
  junctions.coord = list.juncs[[t.names[1]]][,-1]
  for (i in 2:length(t.names)){
    junctions.coord = rbind(junctions.coord, list.juncs[[t.names[i]]][,-1])
  }
  junctions.coord = unique(junctions.coord)
  junctions.coord = junctions.coord[order(junctions.coord$e.beg, junctions.coord$e.end),]
  junctions.ind = matrix(0, nrow = gene$nforms, ncol = nrow(junctions.coord))
  for (i in 1:length(list.juncs)){
    junctions.ind[i, match(list.juncs[[t.names[i]]]$ID, ac(junctions.coord$ID)) ] = 1
  }
  colnames(junctions.ind) = ac(junctions.coord$ID)
  return(list(junctions.ind = junctions.ind, junctions.coord = junctions.coord))
}

MapJunctionCounts <- function(gene, counts.tab){
  stop("Function MapJunctionCounts not implemented")
}

#######################################
#######     Transcript merging
#######   

### From an ensembl gene,
### Merge the transcripts where the distance is less than
### x% of the average length between the 2 transcripts

ReduceTranscriptSpace <- function(gene, perc.min = 0.05, dist.min = NULL){
  if (gene$nforms <= 1){
    return(gene)
  }
  t.l = transcripts.length(gene)
  table.dists = dist.bp.ensembl(gene, table.result = TRUE)
  table.dists$perc = table.dists$distance / 
    pmax(t.l[match(table.dists$ENST.1 , colnames(t.l))] ,
         t.l[match(table.dists$ENST.2 , colnames(t.l))])
  table.dists = table.dists[order(table.dists$perc), ]
  gene.res = gene
  ##merge the nearer transcripts
  while (table.dists$perc[1] < perc.min )
    {
      ct.merge = c(as.character(table.dists$ENST.1[1]),
        as.character(table.dists$ENST.2[1]))
      gene.res = merge.transcripts(gene.res, ct.merge)
      if (gene.res$nforms == 1){
        break
      }
      t.l = transcripts.length(gene.res)
      table.dists = dist.bp.ensembl(gene.res, table.result = TRUE)
      table.dists$perc = table.dists$distance / 
        pmax(t.l[match(table.dists$ENST.1 , colnames(t.l))] ,
             t.l[match(table.dists$ENST.2 , colnames(t.l))])
      table.dists = table.dists[order(table.dists$perc), ]
    }
  return(gene.res)
}

###
###  Utility function to merge transcripts
###  we add a merge component to the gene list
merge.transcripts <- function(gene, trans.list){
  trans.list = unique(trans.list)
  #cat("merge " , trans.list, "\n")
  t.exists = trans.list %in% gene$names
  if (any(!t.exists)){
    warning(paste("Some transcripts do not exist in gene" ,
                  gene$name, ":\n",
                  paste(trans.list[!t.exists], sep = "" , collapse = "-"),"\n" ))
    trans.list = trans.list[t.exists]
  }
  if (length(trans.list) <=1){
    return(gene)
  }
  gene.res = gene
  if (is.null(gene.res$merges)){
    new.name = sub("ENST00", "ENST-M", sort(trans.list)[1])
    gene.res$merges[[new.name]] = trans.list
  }else{
    M.inds = grep("ENST-M", trans.list)
    if (length(M.inds) > 0){
      new.name = sort(trans.list[M.inds])[1]
      M.in.M = match(trans.list[M.inds] , names(gene.res$merges))
      all.trans = c(unlist(gene.res$merges[M.in.M]), trans.list[-M.inds])
      names(all.trans) = NULL
      gene.res$merges = gene.res$merges[-M.in.M]
      gene.res$merges[[new.name]] = all.trans
    }else{
      new.name = sub("ENST00", "ENST-M", sort(trans.list)[1])
      gene.res$merges[[new.name]] = trans.list      
    }
  }

  m.trans = match(trans.list, gene$names)
  m.Itrans = match(trans.list, row.names(gene$I.big))
  new.Iline = as.integer(apply(gene$I.big[m.Itrans,], 2, sum) > 0)
  new.Iline = as.data.frame(matrix(new.Iline, nrow = 1,
    dimnames = list(new.name, colnames(gene$I.big))))
  new.nforms = gene$nforms - length(trans.list) + 1
  if (gene$njunctions > 0){
    if (gene$njunctions > 1){
      new.Jline = as.integer(apply(gene$junctions.ind[m.Itrans,],2, sum) >0)
      gene.res$junctions.ind = gene.res$junctions.ind[-m.Itrans,]
    }else{
      new.Jline = as.integer(sum(gene$junctions.ind[m.Itrans]) > 0)
      gene.res$junctions.ind = gene.res$junctions.ind[-m.Itrans]
    }
  }
  gene.res$nforms = new.nforms
  gene.res$names = gene.res$names[-m.trans]
  gene.res$I.big = gene.res$I.big[-m.Itrans,]
  if (new.nforms == 1){
    gene.res$I.big = matrix(new.Iline, nrow = 1,
      dimnames = list(new.name, colnames(gene$I.big)))
    if (gene$njunctions >0){
      gene.res$junctions.ind = matrix(new.Jline, nrow = 1)
    }
  }else{
    gene.res$I.big = rbind(new.Iline, gene.res$I.big)
    if (gene$njunctions == 1){
      gene.res$junctions.ind = c(new.Jline, gene.res$junctions.ind)
    }
    if (gene$njunctions > 1){
      gene.res$junctions.ind = rbind(new.Jline, gene.res$junctions.ind)
    }
  }
  gene.res$names = c(new.name , gene.res$names)
  return(gene.res)
}

####################   End transcript merging

##########
######  Detecting overlaps between genes

###
### The overlaps are data frames with begin and end
### positions and exon corresponding to the exon number
### considered
SubstractOverlapEnsembl <- function(gene, overlaps){
  ###
  slices = NULL
  exon.overlapped = unique(overlaps$exon)
  for (cex in exon.overlapped){
    curslice = slice.exon(gene$exons.pos[cex,], subset(overlaps, exon == cex))
    slices = rbind(slices, curslice)
  }
  slices$subexon = factor(slices$subexon, exclude = NULL)
  exons.pos = gene$exons.pos
  I.big = gene$I.big
  exons.pos$subexon.old = exons.pos$subexon
  subex.overlapped = levels(slices$subexon)
  slices$subexon.old = slices$subexon
  slices$subexon.old = unlist(tapply(slices$subexon,slices$subexon, function(x){
    paste(as.character(x[1]), 1:length(x), sep = ".") }))
  tmp = tapply(slices$subexon,slices$subexon, function(x){
    matrix(rep(I.big[, x[1]], length(x)), ncol = length(x))})
  exons.duplication = NULL
  for (i in 1:length(tmp)){
    exons.duplication = cbind(exons.duplication, tmp[[i]])
  }
  ##Take out non interesting exons
  exons.pos = subset(exons.pos, ! subexon %in% subex.overlapped)
  exons.pos = rbind(exons.pos, data.frame(slices, dist.next = 0))
  new.sort = order(order(exons.pos$begin))
  exons.pos = exons.pos[order(exons.pos$begin),]
  ##The same for the Indicator matrix
  c.degage = which(! colnames(I.big) %in% subex.overlapped)
  I.big = I.big
  I.big = I.big[, c.degage]
  I.big = cbind(I.big , exons.duplication)
  I.big = I.big[, new.sort]
  
  exons.ind = with(exons.pos, begin != 0 & end != 0)
  exons.pos = exons.pos[exons.ind,]
  if (is.null(dim(I.big))){
    I.big = matrix(I.big, ncol = length(I.big))
  }
  #cat(dim(I.big), length(I.big), exons.ind, "1-\n")
  I.big = I.big[,which(exons.ind)]
  if (length(I.big)> 0 & is.null(dim(I.big))){
    I.big = matrix(I.big, ncol = length(I.big))
  }
  ###R recast automatically matrices to vectors
  ###if there is 1 element...
  nexons.new = nrow(exons.pos)
  if (nexons.new <= 1){
    exons.pos$dist.next = rep(NA, nexons.new)
  }else{
    exons.pos$dist.next = c(exons.pos$begin[2:nexons.new]-exons.pos$end[1:(nexons.new-1)],NA)
    exons.pos$subexon = paste("E", 1:nexons.new , sep = "")
    #cat(dim(I.big),"ffff\n")
    colnames(I.big) = paste("E", 1:nexons.new , sep = "")
  }
  gene$nexons = nexons.new
  gene$exons.pos = exons.pos
  gene$I.big = I.big
  return(gene)
}

slice.exon <- function(exon.desc, overlaps){
  overlaps = overlaps[order(overlaps$begin),]
  br = sort(c(unique(c(exon.desc$begin, overlaps$begin)), unique(c(exon.desc$end, overlaps$end))))
  slices = NULL
  for (i in 1:(length(br)-1)){
    if (all(br[i] != overlaps$begin)){
      slices = rbind(slices, data.frame(begin = br[i], end = br[i+1],
        length = br[i+1] - br[i] + 1))
    }
  }
  if(is.null(slices)){
    ###The whole exon will be erased
    return(data.frame(subexon = exon.desc$subexon, begin = 0, end = 0,length = 0))
  }
  else{
    return(data.frame(subexon = exon.desc$subexon, slices))
  }
}

###
### A function to substract positions which are shared by multiple genes
### We use the big table of exons coordinates and of genes infos in order
### to proceed by chromosomes

GetIntersectingGenes <- function(genes.info, genes.coord){
  fac.genes = factor(genes.coord$gene, exclude = NULL)
  genes.borders = tapply(1:nrow(genes.coord),fac.genes , function(x){ return(range(c(genes.coord$begin[x], genes.coord$end[x])))})
  genes.borders = as.data.frame(t(sapply(genes.borders, c)))
  names(genes.borders) = c("begin", "end")
  infos.borders = merge(genes.info, genes.borders, by.x = 1, by.y = 0)
  infos.borders = infos.borders[order(infos.borders$chr, infos.borders$begin),]
  ngenes = nrow(infos.borders)
  dist.next = infos.borders$end[1:(ngenes-1)] -
    infos.borders$begin[2:ngenes] + 1
  dist.next[infos.borders$chr[2:ngenes] != infos.borders$chr[1:(ngenes-1)]] = NA
  infos.borders$dist.next = c(dist.next, NA)
  overlap.indexes = which(dist.next > 0)
  ## Check more carefully on all the indexes
  ##
  genes.with.intersect = NULL
  #cat("Testing", length(overlap.indexes), " genes\n")
  for (ind in overlap.indexes){
    ###
    curgene = as.character(infos.borders$gene[ind])
    curchr = as.character(infos.borders$chr[ind])
    ##Get the genes that could overlap
    ##that is, compute the distance from ind+2 to end of chromosome
    tocheck = 1:ngenes %in% (ind+1):ngenes & curchr == infos.borders$chr
    with.overlap = which(infos.borders$end[ind] - infos.borders$begin[tocheck] + 1 > 0) + ind
    if (length(with.overlap) > 1){
      cat("Gene ", curgene," avec ", length(with.overlap),  " overlaps\n")
      cat(as.character(infos.borders$gene[with.overlap]), "\n", sep = "-")
      cat("Number " ,which(ind == overlap.indexes), "of", length(overlap.indexes),"\n")
    }
    cg.exons = subset(genes.coord, gene == curgene)
    #cat("Testing ",curgene, "against",
    #    as.character(infos.borders$gene[with.overlap]),"\n")
    for (compg in with.overlap){
      gene.right = as.character(infos.borders$gene[compg])
      cur.intersect = GetIntersectPositions(cg.exons,
        subset(genes.coord, gene == gene.right),
        n1 = curgene, n2 = gene.right)
      #cat(gene.right, ": ", nrow(cur.intersect), "intersections\n" )
      genes.with.intersect = rbind(genes.with.intersect, cur.intersect)
    }
  }
  return(genes.with.intersect)
}

##
##Substract the coordinate of coord2 intersecting with coord1
##and return a new exon coordinate object.
GetIntersectPositions <- function(coord1, coord2, n1 = "a", n2 = "b"){
  ##test each exon against the others
  overlap.comb = NULL
  start = which(coord1$end >= min(coord2$begin))
  if (length(start) == 0){
    return(NULL)
  }
  start = min(start)
  for (i in start:nrow(coord1)){
    possible = which(coord1$end[i] >= coord2$begin)
    if (length(possible) == 0){
      next
    }
    overlaps = overlap.boxes(coord1[i,], coord2[possible, ],
      incr = min(possible)-1, ind1 = i)
    overlap.comb = rbind(overlap.comb, overlaps)
  }
  if (!is.null(overlap.comb)){
    overlap.comb =data.frame(gene1 = n1 , gene2 = n2,overlap.comb)
  }
  return(overlap.comb)
}

###
### get the begin and end position of
### the overlapping boxes
overlap.boxes <- function(box1, boxes2, incr = 0, ind1 = 1){
  o.begin = pmax(box1$begin, boxes2$begin)
  o.end   = pmin(box1$end, boxes2$end)
  ind = which(o.begin <= o.end)
  if (length(ind)> 0){
    return(data.frame(num1 = ind1, num2= ind + incr,begin= o.begin[ind],end = o.end[ind]))
  }
  else{
    return(NULL)
  }
}

overlap.sizes <- function(box1, boxes2)
{
  dists = box1$end - boxes2$begin
  corr1 = pmax(box1$begin - boxes2$begin ,0)
  corr2 = pmax(box1$end - boxes2$end, 0)
  return(dists - corr1 - corr2)
}


######
################  End intersecting positions  ######################

#################################################################
###
###   Methods for analysis and utilities
###

##
## lengths of the transcripts
##
transcripts.length <- function(gene){
  t.l = as.vector(gene$exons.pos$length) %*% t(as.matrix(gene$I.big))
  return(t.l)
}


####
#### transcripts bp distances
####
dist.bp.ensembl <- function(gene, table.result = FALSE){
  warning("Computing distance between ensembl genes, not correcting on junctions hits")
  all.dist = dist(t(t(gene$I.big) * gene$exons.pos$length),
    method = "manhattan", diag = TRUE)
  if (table.result){
    all.dist = as.matrix(all.dist)
    all.dist[upper.tri(all.dist, diag = TRUE)] = NA
    all.dist = na.omit(as.data.frame(as.table(all.dist)))
    names(all.dist) = c("ENST.1", "ENST.2", "distance")
  }
  return(all.dist)
}

####
#### Reducing the I matrix on the smallest group of exons
#### that differentiate between the multiple isoforms.
#### 
reduce.ind.exons <- function(submat){
  submat = as.matrix(submat)
  colkeep = which(!is.na(submat[1,]))
  mat.ind = t(submat[,colkeep])
  ###Identical rows are now merged
  bases = 2^(1:ncol(mat.ind))
  ### We compute a F matrix which does the merging by acting on I
  ###i.e. I %*% F ou exons.length %*% F will give the new exons.
  ind = apply(t(mat.ind)*bases, 2, sum)
  trace = unique(ind)
  F = matrix(0, nrow = length(ind), ncol = length(trace))
  for (i in 1:length(trace)){
    ccomb = trace[i]
    F[,i] = as.numeric(ind == ccomb)
  }
  new.mat = t(mat.ind) %*% F
  res = list(F = F, I.red = new.mat)
  return(res)
}


###
###Get exons numbers from a gene and coordinates
###The coordinates are a 3 columns vector
Exon.from.coord <- function(gene, begin, end){
  an = as.numeric
  e.p = gene$exons.pos
  exon.sel = which(e.p$begin >= begin &
    e.p$end <= end)
  return(exon.sel)
}

###
###Get the Ratio between the exon number exon.number and
###the exon.out exons list in gene.
Exon.Ratio <- function(gene, counts, exon.number, exon.out = -exon.number){
  n.hits = gene$exons.pos$n.uniq.hits
  n.junc.hit = gene$junctions.coord$length
  n.hits.out = sum(n.hits[exon.out])
  # + sum(n.junc.hit)
  n.hits.in = sum(n.hits[exon.number])
  counts.out = sum(counts$counts.exons[exon.out])
  #sum(counts$counts.junctions)
  counts.in = sum(counts$counts.exons[exon.number])
  in.NE = (counts.in / n.hits.in)
  out.NE = counts.out / n.hits.out
  res = list( exon.NE = in.NE, exon.out.NE = out.NE,
    ratio.AS = in.NE / (counts.out / n.hits.out - counts.in / n.hits.in))
  return(res)
}



##
##Transform an Exon pos table in an abstract exons one
###We still have a problem with the definition of unique
###hits on the connex junctions !!!
ToAbsExons <- function(exons.pos, counts = NULL, gene = NULL, read.length = 27){
  ac = as.character
  dist.next = na.omit(exons.pos$dist.next)
  nex = nrow(exons.pos)
  ex.nums = as.numeric(sub("E", "", ac(exons.pos$subexon)))
  abs.ex = factor(c(1,
    cumsum(!(dist.next == 1 & (ex.nums[-nex]+1) == ex.nums[-1])) + 1))
  if(length(abs.ex) > nex){
    abs.ex = abs.ex[-length(abs.ex)]
  }
  if(is.null(counts) | is.null(gene)){
    absexons.pos = by(exons.pos, abs.ex, function(x){
      with(x,data.frame(subexon = subexon[1], begin = min(begin), end = max(end),
                        length = sum(length),
                        n.hits = sum(n.hits),
                        n.uniq.hits = sum(n.uniq.hits),
                        dist.next = max(dist.next) ))
  })}else{
    absexons.pos = by(exons.pos, abs.ex, function(x){
      ex.num = which(as.character(gene$exons.pos$subexon)
        %in% as.character(x$subexon))
      cur.count = sum(counts$counts.exons[ex.num])
      l.add = 0
      n.h.add = 0
      n.u.h.add = 0
      if (gene$njunctions > 0){
        j.c = gene$junctions.coord[,c("e.beg", "e.end", "length")]
        j.c$num = 1:nrow(j.c)
        junc.keep = merge(j.c, data.frame(e.beg = ex.num[1:(length(ex.num)-1)],
          e.end = ex.num[2:length(ex.num)]))
        if(length(junc.keep) > 0){
          cur.count = cur.count + sum(counts$counts.junctions[junc.keep$num])
          l.add = l.add + sum(junc.keep$length)
          n.h.add = n.h.add + sum(junc.keep$length-read.length)
          n.u.h.add = n.h.add
        }
      }
      with(x,data.frame(subexon = subexon[1], begin = min(begin), end = max(end),
                        length = sum(length) + l.add,
                        n.hits = sum(n.hits) + n.h.add,
                        n.uniq.hits = sum(n.uniq.hits) + n.u.h.add,
                        dist.next = max(dist.next),
                        counts = cur.count,
                        exons = paste(ex.num, collapse = "-")
                        ))
    })}
  res.absexons=  absexons.pos[[1]]
  for( i in 2:length(absexons.pos)){
    res.absexons = rbind(res.absexons, absexons.pos[[i]])
  }
  res.absexons$length = res.absexons$end - res.absexons$begin + 1
  res.absexons$dist.next = c(res.absexons$end[2:nrow(res.absexons)]- res.absexons$begin[1:(nrow(res.absexons)-1)]+1, NA)
  return(res.absexons)
}


####
####  Return a comprehensive annotation from a gene structure
####

###
###  From a gene, output a table of the relative AS types
###  between all transcripts
###  all the types are : _ Mutually Exclusive    ME 
###                      _ cassette exon         CE
###                      _ alternative donor     AD
###                      _ alternative acceptor  AA
###                      _ alternative promoter  AP
###                      _ alternative terminal  AT

### The dumb way for the moment (not extensively tested)
GetAStypes <- function(gene){
  AS.types = c("ME", "CE", "AD", "AA", "AP", "AT")
  nexons = gene$nexons
  nforms = gene$nforms
  I.mat = gene$I.big
  fnames = rownames(I.mat)
  exons.bound = gene$exons.pos$dist.next == 1
  res = data.frame(form.1 = NA, form.2 = NA , AS.type = "unknown")
  for (i in 1:(nforms-1)){
    for (j in (i+1):nforms){
      cur.as = c() 
      ue = which(I.mat[i,] + I.mat[j,] > 0)
      nex = length(ue)
      diff.exons = I.mat[i,] - I.mat[j,]
      ##smooth the signal according to the neighbourhood
      if (all(diff.exons[ue] != 0)){
        cur.as = c("ALL")
        next
      }
      k = 2
      if (diff.exons[ue[1]] != 0){
        cur.as = c(cur.as, "AP")
        while( diff.exons[ue[k-1]] == diff.exons[ue[k]]){
          k = k + 1
        }
      }
      if (diff.exons[nex] != 0){
        cur.as = c(cur.as, "AT")
        while (diff.exons[ue[nex]] == diff.exons[ue[nex-1]]){
          nex = nex - 1
        }
      }
      ###now all the internal exons
      while(k < nex){
        l = 1
        if (diff.exons[ue[k]] != 0){
          #all the next exons until concording
          l = k+1
          while (diff.exons[ue[l]] != 0 & l < (nex-1)){
            l = l+1
          }
          if (all(c(-1,1) %in% diff.exons[ue[k:(l-1)]])){
            cur.as = c(cur.as, "ME")
          }else{
            if ((ue[k-1] == ue[k]-1) & exons.bound[ue[k-1]]){
              cur.as = c(cur.as, "AA")
            } else {
              if ((ue[l] == ue[l-1]+1) & exons.bound[ue[l-1]]){
                cur.as = c(cur.as, "AD")
              } else {
                cur.as = c(cur.as, "CE")
              }
            }
          }
        }
        k = k+l
      }
      res = rbind(res , data.frame(form.1 = fnames[i], form.2 = fnames[j],
        AS.type = paste(sort(cur.as), sep ="-" )))
    }
  }
  return(res)
}

##
## The EM algorithm wrapper (kept separated for easier integration of more
## complex cases which are to come).
## 
EstimateProportions <- function(gene, counts, read.length = 27,...){
  if (!is.null(gene$exons.pos$n.uniq.hits)){
    l.exons = gene$exons.pos$n.uniq.hits
  }else{
    l.exons = pmax( gene$exons.pos$length -read.length + 1,0) 
  }
  if (gene$njunctions > 0){
    Indicator = t(cbind(gene$I.big, gene$junctions.ind))
    l.exons = c(l.exons , pmax(gene$junctions.coord$length - read.length + 1, 0))
  }else{
    Indicator = t(gene$I.big)
  }
  counts = c(counts$counts.exons, counts$counts.junctions)
  gene4EM = create.gene(Indicator, l.exons)
  gene.estimated = em.gene.estimate.vect(gene4EM, counts,...)
  ###We should add at least the transcript names on the proportions
  ###here.
  return(gene.estimated)
}


##Simple function that return a matrix of 1 for border exons
##from a gene object
GetBorderExon <- function(gene){
  I.mat = gene$I.big
  nexons = gene$nexons
  dist.next = na.omit(gene$exons.pos$dist.next)
  abs.ex = factor(c(1, cumsum(dist.next != 1) + 1))
  Ex.Nul = matrix(0, ncol = ncol(I.mat), nrow = nrow(I.mat))
  for (i in 1:nrow(I.mat)){
    borders = range(which(I.mat[i,] == 1))
    ex.sID = abs.ex %in% abs.ex[borders]
    ex.sID = ex.sID &  1:nexons >= borders[1] & 1:nexons <= borders[2] & I.mat[i,] == 1
    Ex.Nul[i, ex.sID ] = 1
  }
  return(Ex.Nul)
}

###
### The POEM tool, kept as a different function from EstimateProportions
### to integrate more complex cases in the estimation

POEM.gene <- function(gene, counts, read.length = 27,
                      data.type = "SR", scenario = "basic" ,...){
  if (data.type != "SR"){
    stop("POEM.gene only implemented for single reads now\n")
  }
  # Needs to add transcripts information here.
  EM.res = EstimateProportions(gene, counts, read.length = read.length,...)
  EM.res$p.estim
  names(EM.res$p.estim) = gene$names
  return(EM.res)
}


##
## A function to run the proportion estimations on the genes with
## cutted first and last exons
## NOT extensively tested yet
## The outliers.pval option allows to wipe out the exons which have a
## significantly different expression than the baseline.
##
POEM.exons.cut <- function(gene, counts, read.length = 27,
                           outliers.pval = 0, exon2keep = NULL,...){
  counts.2est = counts
  gene.new = gene
  ###
  ### First a cleaning on the exons which seem to be to different
  ### from baseline expression
  ###
  exons.signif.out = NULL
  if (outliers.pval > 0){
    ne.exons = counts$counts.exons / gene$exons.pos$n.uniq
    med.exp = median(ne.exons, na.rm = TRUE)
    pval.exons = sapply(1:length(ne.exons) , function(x) {
      min(ppois(counts$counts.exons[x], med.exp * gene$exons.pos$n.uniq[x]),
        ppois(counts$counts.exons[x], med.exp * gene$exons.pos$n.uniq[x], lower.tail = FALSE))
    })
    exons.signif.out = which(pval.exons <= outliers.pval)
    if (!is.null(exon2keep)){
      exons.signif.out = exons.signif.out[!(exons.signif.out %in% exon2keep)]
    }
  }
  exons.border = GetBorderExon(gene)
  gene.new$I.big = gene$I.big - exons.border
  counts.2est$counts.exons[apply(gene.new$I.big == 0, 2, all)] = 0 
  if (gene$njunctions > 0){
    for (i in 1:nrow(gene$I.big)){
      ex.tmp = which(exons.border[i,] == 1)
      if (!is.null(exons.signif.out)){
        ex.tmp = union(ex.tmp , exons.signif.out)
      }
      j.2wipe = ex.tmp[ex.tmp == c(ex.tmp[-1]-1, NA)]
      j.2wipe = j.2wipe[is.finite(j.2wipe)]
      col.junc = which(gene$junctions.coord$e.beg %in% j.2wipe)
      #cat("We wipe ", col.junc, " from ", gene$name ,"\n")
      if (gene$njunctions > 1){
        gene.new$junctions.ind[i,col.junc] = 0
      } else{
        gene.new$junctions.ind[i] = 0
      }
    }
    if (gene$njunctions > 1){
      counts.2est$counts.junctions[apply(gene.new$junctions.ind == 0, 2, all)] = 0
    }else{
      if (all(gene.new$junctions.ind == 0)){
        counts.2est$counts.junctions = 0
      }
    }
  }
  #
  return(EstimateProportions(gene.new, counts.2est, read.length = read.length, ...))
}


###
### From a gene model, generate the expected counts on junctions,
### (note that we do not have the information of the number of unique
### hits yet)
ExpectedJunctionsCounts <- function(gene, EM.res, read.length = 27){
  J.info = GenerateJunctionMatrix(gene)
  n.junc = apply(J.info$junctions.ind, 1,sum)
  l.forms = EM.res$sf
  i.junc = (n.junc != 0) & (l.forms != 0)
  J.mat = J.info$junctions.ind[i.junc,]/ n.junc[i.junc]
  c.obs = EM.res$counts.estim
  c.junc = c.obs[i.junc] * n.junc[i.junc] * read.length / l.forms[i.junc]
  if (sum(i.junc)>1){
    junc.exp = as.data.frame(apply(c.junc * J.mat, 2, sum))
  }else{
    junc.exp = as.data.frame(J.mat * c.junc)
  }
  names(junc.exp) = "Exp.count"
  res.coord = merge(J.info$junctions.coord, junc.exp , by.x = "ID", by.y = 0)
  res.coord$gene.ID = gene$name
  res.coord$chr = gene$chr
  nc = ncol(res.coord)
  res.coord = res.coord[, c(nc-1, nc, 1:(nc-2))]
  res.coord$P.count.sup.1 = ppois(1,lambda = res.coord$Exp.count, lower.tail = FALSE)
  return(res.coord)
}






#################################################################
###
###  Output functions
###
#################################################################


####
####  Utility function to write a gene to a gff file
####  the estimated counts or probas are added in the score
####  column but must contain the transcript names
####  

print.GeneStructure <- function(gene){
  warning("Method print not yet implemented for GeneStructure\nPrinting raw information")
  print(gene)
}



gene2gff <- function(gene, file, counts = NULL, probas = NULL,
                     note = NULL, feat.suffix = "", min.proba = 0){
  chr = gene$chr
  gene.pos = range(c(gene$exons.pos$begin, gene$exons.pos$end))
  gene.name = gene$name
  drtn = gene$direction
  attr = paste("ID=", gene.name, ";ensembl_id=", gene.name, sep = "")
  if (!is.null(note)){
    attr = paste(attr, ";Note=", note, sep = "")
  }
  ####We could add information about the total and others in the note.
  score = "."
  frame = "."
  seq = paste("POEM", feat.suffix , sep = "")
  writeLines(paste(chr, seq, "gene", gene.pos[1], gene.pos[2],
                   score, drtn, frame, attr, sep = "\t"),
             file)
  #list of transcripts to write
  #We could put the condition on min proba here
  trans2write = gene$names
  if (!is.null(probas) & all(trans2write %in% names(probas))){
    t.o = match(trans2write, names(probas))
    t.dec.o = order(probas[t.o], decreasing = TRUE)
    trans2write = trans2write[t.dec.o]
  }
  
  for (curtrans in trans2write){
    if (is.null(row.names(gene$I.big))){
      Irow = which(curtrans == gene$names)
    }else{
      Irow = which(curtrans == row.names(gene$I.big))
    }
    Iline = which(gene$I.big[Irow,] == 1)
    if (length(Iline) == 0){
      next
    }
    coord = range(unlist(gene$exons.pos[Iline,c("begin","end")]))
    attr = paste("ID=", curtrans,";Parent=", gene.name, ";ensembl_id=",curtrans, sep = "")
    WLine = TRUE
    if(!is.null(probas) & any(names(probas) == curtrans)){
      attr = paste(attr, ";Alias=", round(probas[[curtrans]],digits = 2), sep = "" )
      if (probas[[curtrans]] < min.proba){
        WLine = FALSE
      }
    }

    if (WLine){
      writeLines(paste(chr, seq, "mRNA", coord[1], coord[2],
                       score, drtn, frame, attr, sep = "\t"),
                 file)
    
      ##Every exon
      for (iexons in Iline){
        curexon = as.character(gene$exons.pos$subexon[iexons])
        score = "."
        coord = c(gene$exons.pos$begin[iexons], gene$exons.pos$end[iexons])
        attr = paste("ID=",curtrans,"_",curexon,";Parent=", curtrans, sep = "")
        writeLines(paste(chr, seq, "exon", coord[1], coord[2],
                         score, drtn, frame, attr, sep = "\t"),
                   file)
      }
    }
  }
  
}
