###
###  Set of functions for the statistical analysis of isoform proportions
###  and simulation of transcripts abundances.
###  
###

###  Let's first define a gene model (simple first version)
###  consist of a list with:
###  - number of exons total n
###  - number of forms k
###  - length of the exons l
###  - indicators on the exons used I with n rows and k columns
###  - vector of proportions p
###  - scaling on library size lambda
###  - vector of rates per isoforms : lamb.j
###  - total expected number of counts for the "region" : tot.counts = \sum lamb.j

#### (want to add some information related to the ensembl genes)


#####
#####  TODO list :
##  - Moment estimation with qr method for non square matrices.
##  - C functions and optimisation methods for likelihood maximisation.
##  - integrating combinat function in the package to avoid annoying
##  dependencies.
##  - any type of noise in the simulation part (cf tests for CASI in other
##  directory).
##

##
##Normalizing a vector of probabilities according to a quantile table
NormalizeProbas <- function(probas, q.table){
  lp = length(probas)
  c.p = cumsum(probas)
  n.val = q.table$dist[findInterval(c.p[-lp], q.table$obs)]
  n.val = c(n.val,1) - c(0, n.val)
  return(n.val)
}

##
##The same function, the other way around
CorrectProbas <- function(probas, q.table){
  lp = length(probas)
  c.p = cumsum(probas)
  n.val = q.table$obs[findInterval(c.p[-lp], q.table$dist)]
  n.val = c(n.val,1) - c(0, n.val)
  return(n.val)
}

##
## The cutting distrib parameter is derived from a normalizing
## method
create.gene <- function(indicator, l, prob = rep(1,k)/k, tot.counts = 500,
                        cutting.distrib = NULL){
  gene = list() 
  gene$n = nrow(indicator)
  gene$k = ncol(indicator)
  n = gene$n
  k = gene$k
  if (any(l < 0)){
    stop("Error in create.gene : some provided lengths are negative")
  }
  gene$l = l
  gene$tot.counts = tot.counts 
  if (!is.matrix(indicator) |length(l) != n ){
    stop("Error in create.gene : indicator matrix with bad specification")
  }
  I = indicator
  dimnames(I) = list(exons = paste("E", 1:n, sep = "."),
            forms = paste("F" , 1:k , sep = "."))
  gene$I = I

  gene$p = prob
  ##We also write internally the proportion for each exon
  q = I*l
  sf = apply(q, 2, sum)
  q = t(t(q)/sf)
  ###Testing the quality of q
  if (!all(is.finite(q))){
    warning("Problem in create.gene, q matrix with not finite values\nCheck Isoform input")
  }
  ###If a non uniform cutting law other the transcript length
  ###is provided, we integrate.
  if (!is.null(cutting.distrib)){
    for (i in 1:k){
      q[,i] = CorrectProbas(q[,i], cutting.distrib)
    }
  }
  gene$q = q    ##matrix of exons length ratios within each isoform
  gene$sf = sf  ##length of each isoform
  gene$lambda = tot.counts / sum(gene$p * gene$sf)
  gene$lamb.j = gene$p * gene$sf * gene$lambda
  return(gene)
}

###
### Extract a subset of transcripts from a gene.
### If the gene object contains results from a previous
### estimation, they are discarded.
### 
subset.gene <- function(gene, indexes, drop.empty = TRUE){
#  gene.res = list()
#  gene.res$n = gene$n
#  gene.res$k = length(indexes)
#  gene.res$I = gene$I[, indexes]
#  dimnames(I) = 
#  gene.res$tot.counts = gene$tot.counts
#  gene.res$p = gene$p[indexes]
#  gene.res$q = gene$q[,indexes]
#  gene.res$l = gene$l[indexes]
#  gene.res$sf = gene$sf[indexes]
#  gene.res$lamda = gene$tot.counts / sum(gene.res$p * gene.res$sf)
#  gene.res$lamb.j = gene.res$p * gene.res$sf * gene.res$lambda
  if (!is.null(gene$p.estim)){
    warning("In subset.gene, EM results are not transfered to output\n")
  }
  stopifnot(gene$k >= indexes)
  exons.used = rep(TRUE, gene$n)
  if (drop.empty){
    exons.used = apply(gene$I[,indexes], 1, sum) > 0
  }
  gene.res = create.gene(gene$I[exons.used,indexes],
    gene$l[exons.used], prob = gene$p[indexes])
  return(gene.res)
}


###
###An EM round
###return a list with results (m, p, llike, llike.diff)
em.step <- function(p,q,counts,sf){
  q.t = t(t(q) * sf)
  rsums = as.vector(counts / (q.t %*% p))
  mnew = p * as.vector( t(q.t) %*% rsums )
  ###This could be changed afterward
  mnew = mnew + 1
  lambda = sum(mnew / sf)
  pnew = mnew/ (lambda * sf)
  lamb.j = pnew*sf*lambda
  mu.j = apply(pnew * t(q.t) * lambda, 2, sum)
  llike = -sum(mu.j) + sum(counts * log(mu.j))
  return( list(m = mnew, p = pnew, llike = llike))
}


rdirichlet <- function(n,par){
  # PRODUCE n SAMPLES FROM THE SPECIFIED DIRICHLET DISTRIBUTION
  k=length(par)
  mat=matrix(0,n,k)
  for (i in 1:n){
    sim=rgamma(k,par,1)
    mat[i,]=sim/sum(sim)
  }
  return(mat)
}

###
### This part is too slow, integrate other methods for estimation
### like convex optimisation and moment method for starting points,
### and especially C code for the running part.
em.gene.estimate.vect <- function(gene, counts, mstep = 2000, eps = 1e-6,
                                  true.p = NULL, method = "EM",
                                  random.start = TRUE, nseeds = 10,
                                  trace.return = FALSE, debug.code = FALSE,
                                  goodness.fit = TRUE){
  n = gene$n
  k = gene$k
  l = gene$l
  I = gene$I
  p = gene$p
  p.tr = 0
  ###Init of the parameters and expected values
  lambda = sum(counts)
  gene$tot.counts = lambda
  q = gene$q
  sf = gene$sf
  ##First thing, We control the transcripts with no exons
  q.trace.trans = apply(I, 2, sum)
  d.q = as.matrix(dist(t(I), method = "manhattan"))
  d.q[upper.tri(d.q, diag = TRUE)] = NA
  d.q = na.omit(as.data.frame(as.table(d.q)), stringsAsFactors = FALSE)
  names(d.q) = c("T1", "T2", "dist")
  d.q$T1 = as.character(d.q$T1)
  d.q$T2 = as.character(d.q$T2)
  eq.ctr = which(d.q$dist == 0)
  trans.invisible = NULL
  if (any(q.trace.trans == 0) | length(eq.ctr) > 0){
    trans.invisible = which(q.trace.trans == 0)
    ###If multiple transcripts are strictly equals, we wipe them out
    if (length(eq.ctr) > 0){
      aa = cbind(d.q$T1[eq.ctr], d.q$T2[eq.ctr])
      ##process aa line by line
      while(nrow(aa)>=1){
        cpl = c(aa[1,1], aa[1,2])
        olines = which(( aa[,1] %in% cpl) | (aa[,2] %in% cpl))
        trans.tmp = unique(c(aa[olines,1],aa[olines,2]))
        olines = which(( aa[,1] %in% trans.tmp) | (aa[,2] %in% trans.tmp))
        trans.tmp = unique(c(aa[olines,1],aa[olines,2]))        
        trans.tmp = sapply(strsplit(trans.tmp, ".",fixed =TRUE),
          function(x) as.numeric(x[2]))
        trans.invisible = c(trans.invisible, trans.tmp[-1])
        aa = aa[-olines,]
      }
    }
    q = q[,-trans.invisible]
    I = I[,-trans.invisible]
    p = p[-trans.invisible]
    sf = sf[-trans.invisible]
    k = k - length(trans.invisible)
  }
  if (random.start){
    p.tr = rdirichlet(nseeds, rep(1, k))
    #cat ("Probas start\n")
    #print(p.tr)
  }else{
    p.tr = matrix(1, nrow = nseeds, ncol = k)/ k
  }

  if (method == "moment"){
    return(matrix.gene.estimate(gene, counts, true.p = true.p))
  }
  ##We clean the exons that cannot be seen
  q.trace = apply(q, 1, sum)
  if (any(l == 0) | any(q.trace ==0)){
    ex2wipe = which(l == 0 | q.trace == 0)
    l = l[-ex2wipe]
    n = n - length(ex2wipe)
    q = q[-ex2wipe,]
    I = I[-ex2wipe,]
    counts = counts[-ex2wipe]
  }

  if (debug.code){
    cat("Matrix I:\n")
    print(t(I))
    print(sf)
    print(l)
    print(counts)
  }
  ##Lets aggregate the identical columns for 

  m.tr = 0 ##save the current values of m in a nseeds x nforms matrix
  llike.mat = matrix(0, nrow = mstep, ncol = nseeds)
  niter = -1
  ###Do the EM (vectorized version)
  for (i in 1:mstep){
    param.step = apply(p.tr, 1, function(x) { em.step(as.vector(x), q, counts,sf)})
    m.tr = t(sapply(param.step , function(x) {x$m}))
    p.tr = t(sapply(param.step , function(x) {x$p}))
    llike.mat[i,] = as.vector(sapply(param.step , function(x) {x$llike}))
    #cat("Step", i,"\nProbas:", p.tr, "\nlikelihoods : ", llike.mat[i,], "\n")
    if (i > 1){
      llike.diff = (llike.mat[i,] - llike.mat[i-1,])
      #cat("Step:", i, ", Diffs : " , sum(llike.diff < 0) , "\n")
      llike.diff = max(abs(llike.diff))
      if (llike.diff < eps){
        niter = i
        #cat("Finished after ", niter, "iterations\n")
        llike.mat = llike.mat[1:niter,]
        break 
      }
    }
  }
  ###Taking the model with the best likelihood
  model.win = which.max(llike.mat[nrow(llike.mat),])
  gene$nite = niter
  gene$likelihoods = as.vector(llike.mat[,model.win])
  gene$p.estim = as.vector(p.tr[model.win,])
  gene$counts.estim = as.vector(m.tr[model.win,])
  
  if (!is.null(trans.invisible)){
    gene$p.estim = rep(0, gene$k)
    gene$counts.estim = rep(0, gene$k)
    gene$p.estim[!(1:gene$k %in% trans.invisible)] = as.vector(p.tr[model.win,])
    gene$counts.estim[!(1:gene$k %in% trans.invisible)] = as.vector(m.tr[model.win,])
  }
  
  if (!is.null(true.p)){
    gene$true.p = true.p
    accuracy = list()
    diff = gene$p.estim - true.p
    ord = order(true.p, decreasing = TRUE)
    first = ord[1]
    second = ord[2]
    accuracy$mse = sqrt(sum(diff^2))/ k
    accuracy$tv.mean = sum(abs(diff)) / k
    accuracy$max.e = max(abs(diff))
    accuracy$diff.1st = abs(diff[first])
    accuracy$diff.2nd = abs(diff[second])
    #accuracy$tau = cor(gene$p.estim, true.p, method = "kendall")
    accuracy$niter = niter
    gene$accuracy = unlist(accuracy)
  }
  if (trace.return){
    gene$trace = llike.mat
    gene$all.p = p.tr
    gene$all.counts = m.tr
  }

  if (goodness.fit){
    prop.glob = as.vector(p.tr[model.win,]) * sf
    prop.glob = prop.glob / sum(prop.glob)
    probas.all = q %*% diag(prop.glob)
    exons.prob = apply(probas.all, 1, sum)
    if (sum(counts > 0)){
      gene$quality.score = chisq.test(counts, p = exons.prob)
    }else{
      gene$quality.score = NA
    }
  }
  #print(gene$accuracy, 3)
  return(gene)
}

##
## Estimating the parameters with a simple matrix inversion
## Only working for square matrices of full rank now.
## 
matrix.gene.estimate <- function(gene, counts, true.p = NULL){
  ##first merge equal lines
  q.tmp = (gene$q)
  sf.tmp = (gene$sf)
  k = gene$k
  q.tmp = t(t(q.tmp) * sf.tmp)
  x = solve(q.tmp, counts)
  prob.res = x / sum(x)
  gene$p.estim = prob.res
  if (!is.null(true.p)){
    gene$true.p = true.p
    accuracy = list()
    diff = gene$p.estim - true.p
    ord = order(true.p, decreasing = TRUE)
    first = ord[1]
    second = ord[2]
    accuracy$mse = sqrt(sum(diff^2))/ k
    accuracy$tv.mean = sum(abs(diff)) / k
    accuracy$max.e = max(abs(diff))
    accuracy$diff.1st = abs(diff[first])
    accuracy$diff.2nd = abs(diff[second])
    #accuracy$tau = cor(gene$p.estim, true.p, method = "kendall")
    gene$accuracy = unlist(accuracy)    
  }
  return(gene)
}


###
###The other method, we get the proportions
###directly from the estimated model and use a
### chisq.test method directly
goodness.of.fit <- function(gene, counts){
  l = gene$l
  q = gene$q
  p.estim = gene$p.estim
  q.trace = apply(q, 1, sum)
  if (any(l == 0) | any(q.trace == 0)){
    ex2wipe = which(l == 0 | q.trace == 0)
    l = l[-ex2wipe]
    q = q[-ex2wipe,]
    counts = counts[-ex2wipe]
  }
  #
  prop.glob = gene$p.estim * gene$sf
  prop.glob = prop.glob / sum(prop.glob)
  ## This can be get directly from the q matrix.
  probas.all = q %*% diag(prop.glob) 
  exons.prob = apply(probas.all, 1, sum)
  good.test =  chisq.test( counts, p = exons.prob )
  return( good.test )
}

###
###Compute loglikelihood of a sample
Loglike <- function(gene, counts){
  ##
  if (is.null(gene$p.estim)){
    stop("gene object do not possess a proper proba slot, run EM first.")
  }
  ##put that into EM
  em.res = em.step(gene$p.estim , gene$q, counts, gene$sf)
  return(em.res$llike)
}

###
### Get the expected exons count
###
ExonsExpectedCount <- function(gene, tot.count){
  if (is.null(gene$counts.estim)){
    stop("gene object do not possess a proper exon slot, run EM first.")
  }
  prop.glob = gene$p.estim * gene$sf
  prop.glob = prop.glob / sum(prop.glob)
  ## This can be get directly from the q matrix.
  exp.count.all = gene$q %*% diag(prop.glob * tot.count)
  exons.exp = apply(exp.count.all, 1, sum)
  return(exons.exp)
}


###
### Redo bootstraps estimates on the estimated probas
### we return the full bootstrap sample for later work
### 
###   !!!!!! We do not constraint the total count falling 
###   into the gene when doing the bootstrap (cf Wang & lindsay).
### this is to assess (cf publication from bioinfo on RNA-Seq).
Bootstrap.estimate <- function(gene, nboot = 100,...){
  if (is.null(gene$p.estim)){
    stop("Cannot do bootstrap estimate on a gene with non estimated proportions\n")
  }
  t.c = sum(gene$counts.estim)
  boot.sample = simul.exons.counts(nboot, gene, prob = gene$p.estim,
    tot.counts = t.c)
  boot.prob = t(sapply(1:nboot, function(x) {
    gene.est = em.gene.estimate.vect(gene, boot.sample$exons.count[x,],...)
    return(gene.est$p.estim)
  }))
  colnames(boot.prob) = names(gene$p.estim)
  #stop("function not tested")
  return(boot.prob)
}


gene.ASI.test <- function(gene,x,y=NULL,
                     cell="one", minexon=1,
                     mingene=1, exp=NULL,...){ 
 #check input
  if(mingene <1 || minexon <1)	
    stop("cannot execute function with invalid parameters 
						'mingene or 'minexon '")


  n=gene$nexons
  exons.length = gene$exons.pos$length
  if (!is.null(gene$exons.pos$n.uniq.hits)){
    exons.length = gene$exons.pos$n.uniq.hits
  }
  if (!(cell %in% c("one", "two"))){
    stop("cell parameter must be one or two\n")
  }
  c.x = x$counts.exons
  c.y = NULL
  if (!is.null(y)){
    c.y = y$counts.exons
  }

  if(cell=="one"){
    if(!is.null(y))
      warning("unused parameter y if cell == 'one'")
    ##test the attribute of the gene and the counts
    
   #do CASI analysis    
    if(length(c.x)!=n)
      stop("'x' is of different length than the gene.")    

    # check if mincount options are satisfiable
    if(sum(c.x)<mingene || sum(c.x>minexon)<1){
      warning("Can not execute test on gene with current parameters.
				  The counts in x and/or y are not sufficient.
				  return NA.")
      return(NA)
    }
    valid=c.x>minexon 
    p=exons.length[valid]
    if (sum(valid) <= 1){
      warning("Too few exons are taken into account \nreturn NA")
      return(list(p.value = NA, z.score = NA, statistic = NA))
    }
    test=chisq.test(c.x[valid],p=p/sum(p),...)
    METHOD="CASI Chi-squared test for Alternative Splicing"
    
    #calculate z-scores
    z=rep(NA,n)
    z[valid]=log2(c.x[valid]/p)
    z[valid]=(z[valid]-median(z[valid]))/mad(z[valid])
    
  } #if not this is DASI analysis
  else {
    if (is.null(y)){
      stop("must supply count parameters if cell == two")
    }
    z=rep(NA,n)
    c.x[c.x==0 | c.y==0 ] = c.x[c.x==0 | c.y==0 ] + 1
    c.y[c.x==0 | c.y==0 ] = c.y[c.x==0 | c.y==0 ] + 1
    valid = c.x > minexon | c.y > minexon
    valid2 = c.x > minexon & c.y > minexon
    if (sum(valid2) <= 1){
      warning("Too few exons are taken into account with the thresholds set \nreturn NA")
      return(list(p.value = NA, z.score = NA, statistic = NA))
    }
    
    #check if more than 80% of cells have count > 4
    x.val=c.x[valid] 
    y.val=c.y[valid]
    minimal=length(x.val[x.val>4]) + length(y.val[y.val>4])
    if( (minimal/(2*length(x.val)) ) > 0.8 ){
    	#use hybrid approach because Chi-Square Approximation is valid
    	test = try(fisher.test(cbind(c.x[valid], c.y[valid]), hybrid=TRUE,workspace=2e7, ...))
	
    }else{
    	#if there are more than 1000 counts use sampling based approach
	if( ( (length(x.val) < 7 & sum(x.val,y.val) < 800))  ){
		test = try(fisher.test(cbind(c.x[valid], c.y[valid]), hybrid=TRUE,workspace=2e7, ...))
	}else{
		show("use sampling based approach because too many counts for exact computation")
		test = try(fisher.test(cbind(c.x[valid], c.y[valid]), simulate.p.value = TRUE, B = 100000, ...))
	}
    } 
    
    
   
    z[valid] = log2(c.x[valid] / c.y[valid])
    #check if we can compute the 
    if (mad(z[valid]) == 0){
      warning("not enough exons to compute median absolute deviation")
       z = NA
    }else{
   	 z[valid]=(z[valid]-median(z[valid]))/mad(z[valid])
    }
  }
  # check if test object exists
  if(inherits(test,"try-error"))
        {
             show("use sampling based approach because too many counts for exact computation")
		test = try(fisher.test(cbind(c.x[valid], c.y[valid]), simulate.p.value = TRUE, B = 100000, ...))

        }
  METHOD = paste("DASI test for differential alternative splicing",test$method,sep="\n")
	if(inherits(test,"try-error"))
        {
		show("Test could not be computed, refer to manual and check input format")
	}else{
  		return(structure(list(statistic=test$statistic,parameter = test$parameter, 
                 p.value = test$p.value, z.score=z,method=METHOD,
                 cell=cell, minexon=minexon, mingene=mingene,
                 geneName=gene$name, observed = x, experiment=exp),
            	class = "htest"))
	}
} 

ASI.test <- gene.ASI.test

#####################################################################################
####
####           Simulation part
####

##
##Simulation of exons counts knowing a gene model
##the parameters prob and tot.counts are derived from the
##gene model if not specified
simul.exons.counts <- function(nsim, gene, prob = NULL,
                               tot.counts = NULL){
  if (is.null(tot.counts)){
    tot.counts = gene$tot.counts
  }
  if (is.null(prob)){
    prob = gene$p
  }
  nforms = gene$k
  sf = gene$sf
  lambda = tot.counts / sum(prob * sf)
  ## nsim times nforms matrix of counts
  ## The total number of counts for a gene is also random and subject to Poisson variation
  counts.tot = sapply(1:nforms, function(x){ rpois(nsim, prob[x]*sf[x]*lambda) })
  exons.count = t(sapply(1:nsim, function(x) {
    apply(sapply(1:nforms, function(y) {
      return(rmultinom(1, counts.tot[x,y] , gene$q[,y]))}),
          1 , sum)} ))
  simul = list(total.count = counts.tot , exons.count = exons.count)
  return(simul)
}

###
### The same function, but from a list of genes
### where every genes are supposed to have the 
### same number of forms
simul.exons.counts.gene.list <- function(list.genes, nforms.used = 0, prob, tot.counts){
  num.forms = sapply(list.genes, function(x) x$k)
  forms.prob = list()
  if (nforms.used == 0){
    stopifnot(all(num.forms == length(prob)))
    list.genes.sampled = list.genes
    nform = num.forms[1]
    forms.prob = lapply(num.forms, function(x) return(prob))
  } else {
    form.select = t(sapply(num.forms, function(x) {sample(x, nforms.used)}))
    list.genes.sampled = lapply(1:length(list.genes),
      function(x) return(subset.gene(list.genes[[x]], form.select[x,], drop.empty = FALSE)))
    nform = nforms.used
    forms.prob = lapply(1:length(num.forms),
      function(x) {res = rep(0, num.forms[x]); res[form.select[x,]] = prob ; return(res)})
  }
  nsim = length(list.genes.sampled)
  q.list = lapply(list.genes.sampled, function(x) {qt =x$q ; dimnames(qt) = NULL;  return(qt)})
  #The as.vector cleans the names F.1..F..k and is thus important
  #to avoid random reshuflling
  lengths = t(sapply(list.genes.sampled, function(x) as.vector(x$sf)))
  lambdas = tot.counts / apply(prob * t(lengths), 2 ,sum)
  #total number of counts for each form and each gene
  pois.forms.l = t(prob * t(lengths)) * lambdas
  counts.tot = t(sapply(1:nsim, function(x) {rpois(nform, pois.forms.l[x,])}))
  exons.count = lapply(1:nsim, function(x){
    apply(sapply(1:nform, function(y){
      #cat("gene", x, "and form", y,"\n")
      rmultinom(1, counts.tot[x,y], q.list[[x]][,y])}),
          1, sum)}
    )
  simul = list(total.count = counts.tot, exons.counts = exons.count, prob = forms.prob)
  return(simul)
}


###
### Getting quality of estimation on a fixed number of simulations
### and given proportions of different forms
###
checksimul.old <- function(nsim, gene, prob, tot.counts = NULL, prob.info = FALSE){
  sims = simul.exons.counts(nsim, gene, prob = prob, tot.counts = tot.counts)
  estimates = t(sapply(1:nsim, function(x){
    return(em.gene.estimate(gene, sims$exons.count[x,], true.p = prob)$accuracy)
  }))
  if (prob.info){
    estimates = cbind(matrix(rep(prob, nsim), nrow = nsim, byrow = TRUE),
      estimates)
  }
  return(estimates)
}


###
### Check the quality of estimation on gene with given prob proportions
### and a fixed total count
checksimul.vect <- function(nsim, gene, prob, tot.counts = NULL, prob.info = FALSE,...){
  sims = simul.exons.counts(nsim, gene, prob = prob, tot.counts = tot.counts)
  estimates = t(sapply(1:nsim, function(x){
    return(em.gene.estimate.vect(gene, sims$exons.count[x,], true.p = prob,...)$accuracy)
  }))
  if (prob.info){
    estimates = cbind(matrix(rep(prob, nsim), nrow = nsim, byrow = TRUE),
      estimates)
  }
  return(estimates)
}

###
###  The same function from a gene.list
###  The nforms.used parameter influence the simulation step.
checksimul.vect.gene.list <- function(list.gene, prob, nforms.used = 0,
                                      tot.counts = NULL, prob.info = FALSE,...){
  sims = simul.exons.counts.gene.list(list.gene,nforms.used = nforms.used, prob = prob, tot.counts = tot.counts)
  nsim = length(list.gene)
  estimates = t(sapply(1:nsim, function(x){
    return(em.gene.estimate.vect(list.gene[[x]], sims$exons.count[[x]],
                                 true.p = sims$prob[[x]],...)$accuracy)
  }))
  if (prob.info){
    estimates = cbind(matrix(rep(prob, nsim), nrow = nsim, byrow = TRUE),
      estimates)
  }
  return(estimates)
}


##
##  Handy function to compute quality of estimation on the grid of all possible
##  proportions, with an output on the quality of the estimates
##
simulation.quality <- function(gene, steps = 10, nsim = 100,
                               tot.counts = NULL,...){
  require(combinat, quiet = TRUE)
  nforms = gene$k
  all.probs = t(xsimplex(nforms, steps))/steps
  #n.a.p = paste("p", 1:nforms, sep = ".")
  results = checksimul.vect(nsim, gene, prob = all.probs[1,], tot.counts = tot.counts, prob.info = TRUE, ...)
  for (i in 2:nrow(all.probs)){
    results = rbind(results,
      checksimul.vect(nsim, gene, prob = all.probs[i,],
                 tot.counts = tot.counts, prob.info = TRUE,...))
  }
  #n.a.p = c(n.a.p , names(results), "lambda")
  res.glob = signif(cbind(results, tot.counts = tot.counts), 3)
  #colnames(res.glob) = n.a.p
  return(res.glob)
}


####
#### Updating simulation functions for a given population
#### of known isoforms
####
#### the list.gene.pop argument is a list of gene items 
simulation.quality.ensembl <- function(list.gene.pop, nforms = 2,
                                       steps = 4, nsim = 100,
                                       tot.counts = NULL,...){
  require(combinat, quiet = TRUE)
  num.forms = sapply(list.gene.pop, function(x) return(x$k))
  genes.compatible = which(num.forms >= nforms)
  ##Function trick to avoid 
  ##the copy of too big parameters
  draw.genes <- function(){
    ind.genes = sample(genes.compatible, nsim, replace = TRUE)
    return(list.gene.pop[ind.genes])
    #form.select = t(sapply(ind.genes, function(x) { sample(num.forms[x], nforms) }))
    #this.draw = lapply( 1:nsim,
    #  function(x) return(subset.gene(list.gene.pop[[ind.genes[x]]], form.select[x,])))
    #return(this.draw)
  }
  #Just test for the moment
  all.probs = t(xsimplex(nforms, steps))/steps
  ##We clean the probs on the one which have at least one 0
  ##and divide the set in 2.
  mask.probs = apply(all.probs, 1, prod)
  probs.prod.2compute = unique(mask.probs[mask.probs != 0])
  all.probs.2compute = all.probs[match(probs.prod.2compute, mask.probs),]
  #
  #all.probs = matrix(c(c(1,1,1)/3, c(1,2,1)/4, c(1,1,2)/4, c(2,1,1)/4), ncol = 3 ,
  #  byrow = TRUE)
  results = checksimul.vect.gene.list(draw.genes(), nforms.used = nforms,
    prob = all.probs.2compute[1,], tot.counts = tot.counts, prob.info = TRUE, ...)
  for (i in 2:nrow(all.probs.2compute)){
    results = rbind(results,
      checksimul.vect.gene.list(draw.genes(), nforms.used = nforms ,
                                prob = all.probs.2compute[i,], tot.counts = tot.counts,
                                prob.info = TRUE, ...))
  }
  res.glob = signif(cbind(results, tot.counts = tot.counts, nforms = nforms), 3)
  #colnames(res.glob) = n.a.p
  return(res.glob)
}

##
## Doing multiple CASI simulation at the same time
## This do not look yet at the z scores distributions
check.CASI.simul <- function(list.gene, prob, nforms.used = 0,
                             tot.counts = NULL,prob.info = FALSE, ...){
  sims = simul.exons.counts.gene.list(list.gene,nforms.used = nforms.used, prob = prob, tot.counts = tot.counts)
  nsim = length(list.gene)
  CASI.res = sapply(1:nsim, function(x){
    gene.ASI.test(list.gene[[x]], sims$exons.count[[x]],minexon = 1,...)$p.value
  })
  if (prob.info){
    estimates = cbind(matrix(rep(prob, nsim), nrow = nsim, byrow = TRUE),
      CASI.res)
  }
  return(estimates)
}

###
###A simulation for the CASI of Marcel
simulation.CASI.ensembl <- function(list.gene.pop, nforms =2,
                                    steps = 4 , nsim = 100,
                                    tot.counts=NULL, ...){
  require(combinat, quiet = TRUE)
  num.forms = sapply(list.gene.pop, function(x) return(x$k))
  genes.compatible = which(num.forms >= nforms)
  ##Function trick to avoid 
  ##the copy of too big parameters
  draw.genes <- function(){
    ind.genes = sample(genes.compatible, nsim, replace = TRUE)
    return(list.gene.pop[ind.genes])
    #form.select = t(sapply(ind.genes, function(x) { sample(num.forms[x], nforms) }))
    #this.draw = lapply( 1:nsim,
    #  function(x) return(subset.gene(list.gene.pop[[ind.genes[x]]], form.select[x,])))
    #return(this.draw)
  }
  ###Here we only test on the probas with no 0
  ### and the first one
  all.probs = t(xsimplex(nforms, steps))/steps
  mask.probs = apply(all.probs, 1, prod)
  probs.prod.2compute = unique(mask.probs[mask.probs != 0])
  if(is.null(probs.prod.2compute)){
    stop("Error in gene probabilities cleaning, nothing left")
  }
  all.probs.2compute = all.probs[c(1,match(probs.prod.2compute, mask.probs)),]
  ###Do the simulation for the counts on the genes
  results = check.CASI.simul(draw.genes(), nforms.used = nforms,
    prob = all.probs.2compute[1,], tot.counts = tot.counts, prob.info = TRUE,...)
  for (i in 2:nrow(all.probs.2compute)){
    results = rbind(results,
      check.CASI.simul(draw.genes(), nforms.used = nforms,
                       prob = all.probs.2compute[i,], tot.counts = tot.counts,
                       prob.info = TRUE,...))
  }
  res.glob = cbind(results, tot.counts = tot.counts, nforms = nforms)
  return(res.glob)
}
